# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""High level API for the hashing interface of `model_signing` library.

Hashing is used both for signing and verification and users should ensure that
the same configuration is used in both cases.

The module could also be used to just hash a single model, without signing it:

```python
model_signing.hashing.hash(model_path)
```

This module allows setting up the hashing configuration to a single variable and
then sharing it between signing and verification.

```python
hashing_config = model_signing.hashing.Config().set_ignored_paths(
    paths=["README.md"], ignore_git_paths=True
)

signing_config = (
    model_signing.signing.Config()
    .use_elliptic_key_signer(private_key="key")
    .set_hashing_config(hashing_config)
)

verifying_config = (
    model_signing.verifying.Config()
    .use_elliptic_key_verifier(public_key="key.pub")
    .set_hashing_config(hashing_config)
)
```

The API defined here is stable and backwards compatible.
"""

from collections.abc import Callable, Iterable
import os
import pathlib
import sys
from typing import Literal, Optional, Union

from model_signing import manifest
from model_signing._hashing import hashing
from model_signing._hashing import io
from model_signing._hashing import memory
from model_signing._serialization import file
from model_signing._serialization import file_shard


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


# `TypeAlias` only exists from Python 3.10
# `TypeAlias` is deprecated in Python 3.12 in favor of `type`
if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias


# Type alias to support `os.PathLike`, `str` and `bytes` objects in the API
# When Python 3.12 is the minimum supported version we can use `type`
# When Python 3.11 is the minimum supported version we can use `|`
PathLike: TypeAlias = Union[str, bytes, os.PathLike]


def hash(model_path: PathLike) -> manifest.Manifest:
    """Hashes a model using the default configuration.

    Hashing is the shared part between signing and verification and is also
    expected to be the slowest component. When serializing a model, we need to
    spend time proportional to the model size on disk.

    This method returns a "manifest" of the model. A manifest is a collection of
    every object in the model, paired with the corresponding hash. Currently, we
    consider an object in the model to be either a file or a shard of the file.
    Large models with large files will be hashed much faster when every shard is
    hashed in parallel, at the cost of generating a larger payload for the
    signature. In future releases we could support hashing individual tensors or
    tensor slices for further speed optimizations for very large models.

    Args:
        model_path: The path to the model to hash.

    Returns:
        A manifest of the hashed model.
    """
    return Config().hash(model_path)


class Config:
    """Configuration to use when hashing models.

    Hashing is the shared part between signing and verification and is also
    expected to be the slowest component. When serializing a model, we need to
    spend time proportional to the model size on disk.

    Hashing builds a "manifest" of the model. A manifest is a collection of
    every object in the model, paired with the corresponding hash. Currently, we
    consider an object in the model to be either a file or a shard of the file.
    Large models with large files will be hashed much faster when every shard is
    hashed in parallel, at the cost of generating a larger payload for the
    signature. In future releases we could support hashing individual tensors or
    tensor slices for further speed optimizations for very large models.

    This configuration class supports configuring the hashing granularity. By
    default, we hash at file level granularity.

    This configuration class also supports configuring the hash method used to
    generate the hash for every object in the model. We currently support SHA256
    and BLAKE2, with SHA256 being the default.

    This configuration class also supports configuring which paths from the
    model directory should be ignored. These are files that doesn't impact the
    behavior of the model, or files that won't be distributed with the model. By
    default, only files that are associated with a git repository (`.git`,
    `.gitattributes`, `.gitignore`, etc.) are ignored.
    """

    def __init__(self):
        """Initializes the default configuration for hashing."""
        self._ignored_paths = frozenset()
        self._ignore_git_paths = True
        self.use_file_serialization()

    def hash(self, model_path: PathLike) -> manifest.Manifest:
        """Hashes a model using the current configuration."""
        ignored_paths = [path for path in self._ignored_paths]
        if self._ignore_git_paths:
            ignored_paths.extend(
                [".git/", ".gitattributes", ".github/", ".gitignore"]
            )

        return self._serializer.serialize(
            pathlib.Path(model_path), ignore_paths=ignored_paths
        )

    def _build_stream_hasher(
        self, hashing_algorithm: Literal["sha256", "blake2"] = "sha256"
    ) -> hashing.StreamingHashEngine:
        """Builds a streaming hasher from a constant string.

        Args:
            hashing_algorithm: The hashing algorithm to use.

        Returns:
            An instance of the requested hasher.
        """
        # TODO: Once Python 3.9 support is deprecated revert to using `match`
        if hashing_algorithm == "sha256":
            return memory.SHA256()
        if hashing_algorithm == "blake2":
            return memory.BLAKE2()

        raise ValueError(f"Unsupported hashing method {hashing_algorithm}")

    def _build_file_hasher_factory(
        self,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 1048576,
    ) -> Callable[[pathlib.Path], io.SimpleFileHasher]:
        """Builds the hasher factory for a serialization by file.

        Args:
            hashing_algorithm: The hashing algorithm to use to hash a file.
            chunk_size: The amount of file to read at once. Default is 1MB. A
              special value of 0 signals to attempt to read everything in a
              single call.

        Returns:
            The hasher factory that should be used by the active serialization
            method.
        """

        def _factory(path: pathlib.Path) -> io.SimpleFileHasher:
            hasher = self._build_stream_hasher(hashing_algorithm)
            return io.SimpleFileHasher(path, hasher, chunk_size=chunk_size)

        return _factory

    def _build_sharded_file_hasher_factory(
        self,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 1048576,
        shard_size: int = 1_000_000_000,
    ) -> Callable[[pathlib.Path, int, int], io.ShardedFileHasher]:
        """Builds the hasher factory for a serialization by file shards.

        Args:
            hashing_algorithm: The hashing algorithm to use to hash a shard.
            chunk_size: The amount of file to read at once. Default is 1MB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            shard_size: The size of a file shard. Default is 1 GB.

        Returns:
            The hasher factory that should be used by the active serialization
            method.
        """

        def _factory(
            path: pathlib.Path, start: int, end: int
        ) -> io.ShardedFileHasher:
            hasher = self._build_stream_hasher(hashing_algorithm)
            return io.ShardedFileHasher(
                path,
                hasher,
                start=start,
                end=end,
                chunk_size=chunk_size,
                shard_size=shard_size,
            )

        return _factory

    def use_file_serialization(
        self,
        *,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 1048576,
        max_workers: Optional[int] = None,
        allow_symlinks: bool = False,
    ) -> Self:
        """Configures serialization to build a manifest of (file, hash) pairs.

        The serialization method in this configuration is changed to one where
        every file in the model is paired with its digest and a manifest
        containing all these pairings is being built.

        Args:
            hashing_algorithm: The hashing algorithm to use to hash a file.
            chunk_size: The amount of file to read at once. Default is 1MB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurrent.futures` library to select the best
              value for the current machine.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.

        Returns:
            The new hashing configuration with the new serialization method.
        """
        self._serializer = file.Serializer(
            self._build_file_hasher_factory(hashing_algorithm, chunk_size),
            max_workers=max_workers,
            allow_symlinks=allow_symlinks,
        )
        return self

    def use_shard_serialization(
        self,
        *,
        hashing_algorithm: Literal["sha256", "blake2"] = "sha256",
        chunk_size: int = 1048576,
        shard_size: int = 1_000_000_000,
        max_workers: Optional[int] = None,
        allow_symlinks: bool = False,
    ) -> Self:
        """Configures serialization to build a manifest of (shard, hash) pairs.

        The serialization method in this configuration is changed to one where
        every file in the model is sharded in equal sized shards, every shard is
        paired with its digest and a manifest containing all these pairings is
        being built.

        Args:
            hashing_algorithm: The hashing algorithm to use to hash a shard.
            chunk_size: The amount of file to read at once. Default is 1MB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            shard_size: The size of a file shard. Default is 1 GB.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurrent.futures` library to select the best
              value for the current machine.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.

        Returns:
            The new hashing configuration with the new serialization method.
        """
        self._serializer = file_shard.Serializer(
            self._build_sharded_file_hasher_factory(
                hashing_algorithm, chunk_size, shard_size
            ),
            max_workers=max_workers,
            allow_symlinks=allow_symlinks,
        )
        return self

    def set_ignored_paths(
        self, *, paths: Iterable[PathLike], ignore_git_paths: bool = True
    ) -> Self:
        """Configures the paths to be ignored during serialization of a model.

        If the model is a single file, there are no paths that are ignored. If
        the model is a directory, all paths are considered as relative to the
        model directory, since we never look at files outside of it.

        If an ignored path is a directory, serialization will ignore both the
        path and any of its children.

        Args:
            paths: The paths to ignore.
            ignore_git_paths: Whether to ignore git related paths (default) or
              include them in the signature.

        Returns:
            The new hashing configuration with a new set of ignored paths.
        """
        self._ignored_paths = frozenset({pathlib.Path(p) for p in paths})
        self._ignore_git_paths = ignore_git_paths
        return self



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sigstore based signature, signers and verifiers."""

import pathlib
import sys
from typing import Optional, cast

from google.protobuf import json_format
from sigstore import dsse as sigstore_dsse
from sigstore import models as sigstore_models
from sigstore import oidc as sigstore_oidc
from sigstore import sign as sigstore_signer
from sigstore import verify as sigstore_verifier
from typing_extensions import override

from model_signing._signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class Signature(signing.Signature):
    """Sigstore signature support, wrapping around `sigstore_models.Bundle`."""

    def __init__(self, bundle: sigstore_models.Bundle):
        """Builds an instance of this signature.

        Args:
            bundle: the sigstore bundle (in `bundle_pb.Bundle` format).
        """
        self.bundle = bundle

    @override
    def write(self, path: pathlib.Path) -> None:
        path.write_text(self.bundle.to_json())

    @classmethod
    @override
    def read(cls, path: pathlib.Path) -> Self:
        content = path.read_text()
        return cls(sigstore_models.Bundle.from_json(content))


class Signer(signing.Signer):
    """Signing using Sigstore."""

    def __init__(
        self,
        *,
        oidc_issuer: Optional[str] = None,
        use_ambient_credentials: bool = True,
        use_staging: bool = False,
        identity_token: Optional[str] = None,
    ):
        """Initializes Sigstore signers.

        Needs to set-up a signing context to use the public goods instance and
        machinery for getting an identity token to use in signing.

        Args:
            oidc_issuer: An optional OpenID Connect issuer to use instead of the
              default production one. Only relevant if `use_staging = False`.
              Default is empty, relying on the Sigstore configuration.
            use_ambient_credentials: Use ambient credentials (also known as
              Workload Identity). Default is True. If ambient credentials cannot
              be used (not available, or option disabled), a flow to get signer
              identity via OIDC will start.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
            identity_token: An explicit identity token to use when signing,
              taking precedence over any ambient credential or OAuth workflow.
        """
        if use_staging:
            self._signing_context = sigstore_signer.SigningContext.staging()
            self._issuer = sigstore_oidc.Issuer.staging()
        else:
            self._signing_context = sigstore_signer.SigningContext.production()
            if oidc_issuer is not None:
                self._issuer = sigstore_oidc.Issuer(oidc_issuer)
            else:
                self._issuer = sigstore_oidc.Issuer.production()

        self._use_ambient_credentials = use_ambient_credentials
        self._identity_token = identity_token

    def _get_identity_token(self) -> sigstore_oidc.IdentityToken:
        """Obtains an identity token to use in signing.

        The precedence matches that of sigstore-python:
        1) Explicitly supplied identity token
        2) Ambient credential detected in the environment, if enabled
        3) Interactive OAuth flow
        """
        if self._identity_token:
            return sigstore_oidc.IdentityToken(self._identity_token)
        if self._use_ambient_credentials:
            token = sigstore_oidc.detect_credential()
            if token:
                return sigstore_oidc.IdentityToken(token)

        return self._issuer.identity_token(force_oob=True)

    @override
    def sign(self, payload: signing.Payload) -> Signature:
        # We need to convert from in-toto statement to Sigstore's DSSE
        # version. They both contain the same contents, but there is no way
        # to coerce one type to the other.
        # See also: https://github.com/sigstore/sigstore-python/issues/1076
        statement = sigstore_dsse.Statement(
            json_format.MessageToJson(payload.statement.pb).encode("utf-8")
        )

        token = self._get_identity_token()
        with self._signing_context.signer(token) as signer:
            bundle = signer.sign_dsse(statement)

        return Signature(bundle)


class Verifier(signing.Verifier):
    """Signature verification using Sigstore."""

    def __init__(
        self, *, identity: str, oidc_issuer: str, use_staging: bool = False
    ):
        """Initializes Sigstore verifiers.

        When verifying a signature, we also check an identity policy: the
        certificate must belong to a given "identity", and must be issued by a
        given OpenID Connect issuer.

        Args:
            identity: The expected identity that has signed the model.
            oidc_issuer: The expected OpenID Connect issuer that provided the
              certificate used for the signature.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
        """
        if use_staging:
            self._verifier = sigstore_verifier.Verifier.staging()
        else:
            self._verifier = sigstore_verifier.Verifier.production()

        self._policy = sigstore_verifier.policy.Identity(
            identity=identity, issuer=oidc_issuer
        )

    @override
    def _verify_signed_content(
        self, signature: signing.Signature
    ) -> tuple[str, bytes]:
        # We are guaranteed to only use the local signature type
        signature = cast(Signature, signature)
        bundle = signature.bundle
        return self._verifier.verify_dsse(bundle=bundle, policy=self._policy)



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machinery for computing digests for a single file.

Example usage for `SimpleFileHasher`:
```python
>>> with open("/tmp/file", "w") as f:
...     f.write("abcd")
>>> hasher = SimpleFileHasher("/tmp/file", SHA256())
>>> digest = hasher.compute()
>>> digest.digest_hex
'88d4266fd4e6338d13b845fcf289579d209c897823b9217da3e161936f031589'
```

Example usage for `ShardedFileHasher`, reading only the second part of a file:
```python
>>> with open("/tmp/file", "w") as f:
...     f.write("0123abcd")
>>> hasher = ShardedFileHasher("/tmp/file", SHA256(), start=4, end=8)
>>> digest = hasher.compute()
>>> digest.digest_hex
'88d4266fd4e6338d13b845fcf289579d209c897823b9217da3e161936f031589'
```
"""

import pathlib
from typing import Optional

from typing_extensions import override

from model_signing._hashing import hashing


class FileHasher(hashing.HashEngine):
    """Generic file hash engine.

    This class is intentionally empty (and abstract, via inheritance) to be used
    only as a type annotation (to signal that API expects a hasher capable of
    hashing files, instead of any `HashEngine` instance).
    """


class SimpleFileHasher(FileHasher):
    """Simple file hash engine that computes the digest iteratively.

    To compute the hash of a file, we read the file exactly once, including for
    very large files that don't fit in memory. Files are read in chunks and each
    chunk is passed to the `update` method of an inner
    `hashing.StreamingHashEngine`, instance. This ensures that the file digest
    will not change even if the chunk size changes. As such, we can dynamically
    determine an optimal value for the chunk argument.
    """

    def __init__(
        self,
        file: pathlib.Path,
        content_hasher: hashing.StreamingHashEngine,
        *,
        chunk_size: int = 1_048_576,
        digest_name_override: Optional[str] = None,
    ):
        """Initializes an instance to hash a file with a specific `HashEngine`.

        Args:
            file: The file to hash. Use `set_file` to reset it.
            content_hasher: A `hashing.StreamingHashEngine` instance used to
              compute the digest of the file.
            chunk_size: The amount of file to read at once. Default is 1MB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            digest_name_override: Optional string to allow overriding the
              `digest_name` property to support shorter, standardized names.
        """
        if chunk_size < 0:
            raise ValueError(
                f"Chunk size must be non-negative, got {chunk_size}."
            )

        self._file = file
        self._content_hasher = content_hasher
        self._chunk_size = chunk_size
        self._digest_name_override = digest_name_override

    def set_file(self, file: pathlib.Path) -> None:
        """Redefines the file to be hashed in `compute`.

        Args:
            file: The new file to be hashed.
        """
        self._file = file

    @property
    @override
    def digest_name(self) -> str:
        if self._digest_name_override is not None:
            return self._digest_name_override
        # Since there is no difference between hashing the file with this engine
        # or reading the file in memory and then using the content hasher
        # directly, we must have the same digest_name.
        return self._content_hasher.digest_name

    @override
    def compute(self) -> hashing.Digest:
        self._content_hasher.reset()

        if self._chunk_size == 0:
            with open(self._file, "rb") as f:
                self._content_hasher.update(f.read())
        else:
            with open(self._file, "rb") as f:
                while True:
                    data = f.read(self._chunk_size)
                    if not data:
                        break
                    self._content_hasher.update(data)

        digest = self._content_hasher.compute()
        return hashing.Digest(self.digest_name, digest.digest_value)

    @property
    @override
    def digest_size(self) -> int:
        return self._content_hasher.digest_size


class ShardedFileHasher(SimpleFileHasher):
    """File hash engine that hashes a portion (shard) of the file.

    By invoking this engine in parallel across disjoint shards, we can speed up
    hashing a single file. However, the hash output depends on the shard size.

    It is the responsibility of the user to compose the digests of each shard
    into a single digest for the entire file.
    """

    def __init__(
        self,
        file: pathlib.Path,
        content_hasher: hashing.StreamingHashEngine,
        *,
        start: int,
        end: int,
        chunk_size: int = 1_048_576,
        shard_size: int = 1_000_000_000,
        digest_name_override: Optional[str] = None,
    ):
        """Initializes an instance to hash a file with a specific `HashEngine`.

        Args:
            file: The file to hash. Use `set_file` to reset it.
            content_hasher: A `hashing.HashEngine` instance used to compute the
              digest of the file shard.
            start: The file offset to start reading from. Must be valid. Reset
              with `set_shard`.
            end: The file offset to stop reading at. Must be stricly greater
              than start. The entire shard length must be less than the
              configured `shard_size`. Reset with `set_shard`.
            chunk_size: The amount of file to read at once. Default is 1MB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            shard_size: The size of a file shard. Default is 1 GB.
            digest_name_override: Optional string to allow overriding the
              `digest_name` property to support shorter, standardized names.
        """
        super().__init__(
            file=file,
            content_hasher=content_hasher,
            chunk_size=chunk_size,
            digest_name_override=digest_name_override,
        )

        if shard_size <= 0:
            raise ValueError(
                f"Shard size must be strictly positive, got {shard_size}."
            )
        self.shard_size = shard_size

        self.set_shard(start=start, end=end)

    def set_shard(self, *, start: int, end: int) -> None:
        """Redefines the file shard to be hashed in `compute`.

        Args:
            start: The file offset to start reading from. Must be valid.
            end: The file offset to stop reading at. Must be stricly greater
              than start. The entire shard length must be less than the
              configured `shard_size`.
        """
        if start < 0:
            raise ValueError(
                f"File start offset must be non-negative, got {start}."
            )
        if end <= start:
            raise ValueError(
                "File end offset must be stricly higher that file start offset,"
                f" got {start=}, {end=}."
            )
        read_length = end - start
        if read_length > self.shard_size:
            raise ValueError(
                f"Must not read more than shard_size={self.shard_size}, got"
                f" {read_length}."
            )

        self._start = start
        self._end = end

    @override
    def compute(self) -> hashing.Digest:
        self._content_hasher.reset()

        with open(self._file, "rb") as f:
            f.seek(self._start)
            to_read = self._end - self._start
            if self._chunk_size == 0 or self._chunk_size >= to_read:
                data = f.read(to_read)
                self._content_hasher.update(data)
            else:
                while to_read >= 0:
                    data = f.read(min(self._chunk_size, to_read))
                    if not data:
                        break
                    to_read -= len(data)
                    self._content_hasher.update(data)

        digest = self._content_hasher.compute()
        return hashing.Digest(self.digest_name, digest.digest_value)

    @property
    @override
    def digest_name(self) -> str:
        if self._digest_name_override is not None:
            return self._digest_name_override
        return f"{self._content_hasher.digest_name}-sharded-{self.shard_size}"



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model serializers that operate at file level granularity."""

from collections.abc import Callable, Iterable
import concurrent.futures
import itertools
import pathlib
from typing import Optional

from typing_extensions import override

from model_signing import manifest
from model_signing._hashing import io
from model_signing._serialization import serialization


class Serializer(serialization.Serializer):
    """Model serializer that produces a manifest recording every file.

    Traverses the model directory and creates digests for every file found,
    possibly in parallel.
    """

    def __init__(
        self,
        file_hasher_factory: Callable[[pathlib.Path], io.FileHasher],
        *,
        max_workers: Optional[int] = None,
        allow_symlinks: bool = False,
    ):
        """Initializes an instance to serialize a model with this serializer.

        Args:
            file_hasher_factory: A callable to build the hash engine used to
              hash individual files.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurrent.futures` library.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.
        """
        self._hasher_factory = file_hasher_factory
        self._max_workers = max_workers
        self._allow_symlinks = allow_symlinks

        # Precompute some private values only once by using a mock file hasher.
        # None of the arguments used to build the hasher are used.
        hasher = file_hasher_factory(pathlib.Path())
        self._serialization_description = manifest._FileSerialization(
            hasher.digest_name, self._allow_symlinks
        )

    @override
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ) -> manifest.Manifest:
        """Serializes the model given by the `model_path` argument.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.

        Returns:
            The model's serialized manifest.

        Raises:
            ValueError: The model contains a symbolic link, but the serializer
              was not initialized with `allow_symlinks=True`.
        """
        paths = []
        # TODO: github.com/sigstore/model-transparency/issues/200 - When
        # Python3.12 is the minimum supported version, the glob can be replaced
        # with `pathlib.Path.walk` for a clearer interface, and some speed
        # improvement.
        for path in itertools.chain((model_path,), model_path.glob("**/*")):
            serialization.check_file_or_directory(
                path, allow_symlinks=self._allow_symlinks
            )
            if path.is_file() and not serialization.should_ignore(
                path, ignore_paths
            ):
                paths.append(path)

        manifest_items = []
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self._max_workers
        ) as tpe:
            futures = [
                tpe.submit(self._compute_hash, model_path, path)
                for path in paths
            ]
            for future in concurrent.futures.as_completed(futures):
                manifest_items.append(future.result())

        return manifest.Manifest(
            model_path.name, manifest_items, self._serialization_description
        )

    def _compute_hash(
        self, model_path: pathlib.Path, path: pathlib.Path
    ) -> manifest.FileManifestItem:
        """Produces the manifest item of the file given by `path`.

        Args:
            model_path: The path to the model.
            path: Path to the file in the model, that is currently transformed
              to a manifest item.

        Returns:
            The itemized manifest.
        """
        relative_path = path.relative_to(model_path)
        digest = self._hasher_factory(path).compute()
        return manifest.FileManifestItem(path=relative_path, digest=digest)



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""High level API for the signing interface of `model_signing` library.

The module allows signing a model with a default configuration:

```python
model_signing.signing.sign("finbert", "finbert.sig")
```

The module allows customizing the signing configuration before signing:

```python
model_signing.signing.Config().use_elliptic_key_signer(private_key="key").sign(
    "finbert", "finbert.sig"
)
```

The same signing configuration can be used to sign multiple models:

```python
signing_config = model_signing.signing.Config().use_elliptic_key_signer(
    private_key="key"
)

for model in all_models:
    signing_config.sign(model, f"{model}_sharded.sig")
```

The API defined here is stable and backwards compatible.
"""

from collections.abc import Iterable
import pathlib
import sys
from typing import Optional

from model_signing import hashing
from model_signing._signing import sign_certificate as certificate
from model_signing._signing import sign_ec_key as ec_key
from model_signing._signing import sign_sigstore as sigstore
from model_signing._signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


def sign(model_path: hashing.PathLike, signature_path: hashing.PathLike):
    """Signs a model using the default configuration.

    In this default configuration we sign using Sigstore and the default hashing
    configuration from `model_signing.hashing`.

    The resulting signature is in the Sigstore bundle format.

    Args:
        model_path: the path to the model to sign.
        signature_path: the path of the resulting signature.
    """
    Config().sign(model_path, signature_path)


class Config:
    """Configuration to use when signing models.

    Currently we support signing with Sigstore (public instance and staging
    instance), signing with private keys and signing with signing certificates.
    Other signing modes can be added in the future.
    """

    def __init__(self):
        """Initializes the default configuration for signing."""
        self._hashing_config = hashing.Config()
        self.use_sigstore_signer()

    def sign(
        self, model_path: hashing.PathLike, signature_path: hashing.PathLike
    ):
        """Signs a model using the current configuration.

        Args:
            model_path: The path to the model to sign.
            signature_path: The path of the resulting signature.
        """
        manifest = self._hashing_config.hash(model_path)
        payload = signing.Payload(manifest)
        signature = self._signer.sign(payload)
        signature.write(pathlib.Path(signature_path))

    def set_hashing_config(self, hashing_config: hashing.Config) -> Self:
        """Sets the new configuration for hashing models.

        Args:
            hashing_config: The new hashing configuration.

        Returns:
            The new signing configuration.
        """
        self._hashing_config = hashing_config
        return self

    def use_sigstore_signer(
        self,
        *,
        oidc_issuer: Optional[str] = None,
        use_ambient_credentials: bool = False,
        use_staging: bool = False,
        identity_token: Optional[str] = None,
    ) -> Self:
        """Configures the signing to be performed with Sigstore.

        The signer in this configuration is changed to one that performs signing
        with Sigstore.

        Args:
            oidc_issuer: An optional OpenID Connect issuer to use instead of the
              default production one. Only relevant if `use_staging = False`.
              Default is empty, relying on the Sigstore configuration.
            use_ambient_credentials: Use ambient credentials (also known as
              Workload Identity). Default is False. If ambient credentials
              cannot be used (not available, or option disabled), a flow to get
              signer identity via OIDC will start.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
            identity_token: An explicit identity token to use when signing,
              taking precedence over any ambient credential or OAuth workflow.

        Return:
            The new signing configuration.
        """
        self._signer = sigstore.Signer(
            oidc_issuer=oidc_issuer,
            use_ambient_credentials=use_ambient_credentials,
            use_staging=use_staging,
            identity_token=identity_token,
        )
        return self

    def use_elliptic_key_signer(
        self, *, private_key: hashing.PathLike, password: Optional[str] = None
    ) -> Self:
        """Configures the signing to be performed using elliptic curve keys.

        The signer in this configuration is changed to one that performs signing
        using a private key based on elliptic curve cryptography.

        Args:
            private_key: The path to the private key to use for signing.
            password: An optional password for the key, if encrypted.

        Return:
            The new signing configuration.
        """
        self._signer = ec_key.Signer(pathlib.Path(private_key), password)
        return self

    def use_certificate_signer(
        self,
        *,
        private_key: hashing.PathLike,
        signing_certificate: hashing.PathLike,
        certificate_chain: Iterable[hashing.PathLike],
    ) -> Self:
        """Configures the signing to be performed using signing certificates.

        The signer in this configuration is changed to one that performs signing
        using cryptographic signing certificates.

        Args:
            private_key: The path to the private key to use for signing.
            signing_certificate: The path to the signing certificate.
            certificate_chain: Optional paths to other certificates to establish
              a chain of trust.

        Return:
            The new signing configuration.
        """
        self._signer = certificate.Signer(
            pathlib.Path(private_key),
            pathlib.Path(signing_certificate),
            [pathlib.Path(c) for c in certificate_chain],
        )
        return self



# Copyright 2025 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Signers and verifiers using certificates."""

from collections.abc import Iterable
import logging
import pathlib

import certifi
from cryptography import exceptions
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509 import oid
from OpenSSL import crypto
from sigstore_protobuf_specs.dev.sigstore.bundle import v1 as bundle_pb
from sigstore_protobuf_specs.dev.sigstore.common import v1 as common_pb
from typing_extensions import override

from model_signing._signing import sign_ec_key as ec_key
from model_signing._signing import sign_sigstore_pb as sigstore_pb


logger = logging.getLogger()


class Signer(ec_key.Signer):
    """Signer using certificates."""

    def __init__(
        self,
        private_key_path: pathlib.Path,
        signing_certificate_path: pathlib.Path,
        certificate_chain_paths: Iterable[pathlib.Path],
    ):
        """Initializes the signer with the key, certificate and trust chain.

        Args:
            private_key_path: The path to the PEM encoded private key.
            signing_certificate_path: The path to the signing certificate.
            certificate_chain_paths: Paths to other certificates used to
              establish chain of trust.

        Raises:
            ValueError: Signing certificate's public key does not match the
              private key's public pair.
        """
        super().__init__(private_key_path)
        self._signing_certificate = x509.load_pem_x509_certificate(
            signing_certificate_path.read_bytes()
        )

        public_key_from_key = self._private_key.public_key()
        public_key_from_certificate = self._signing_certificate.public_key()
        if public_key_from_key != public_key_from_certificate:
            raise ValueError(
                "The public key from the certificate does not match "
                "the public key paired with the private key"
            )

        self._trust_chain = x509.load_pem_x509_certificates(
            b"".join([path.read_bytes() for path in certificate_chain_paths])
        )

    @override
    def _get_verification_material(self) -> bundle_pb.VerificationMaterial:
        def _to_protobuf_certificate(certificate):
            return common_pb.X509Certificate(
                raw_bytes=certificate.public_bytes(
                    encoding=serialization.Encoding.DER
                )
            )

        chain = [_to_protobuf_certificate(self._signing_certificate)]
        chain.extend(
            [
                _to_protobuf_certificate(certificate)
                for certificate in self._trust_chain
            ]
        )

        return bundle_pb.VerificationMaterial(
            x509_certificate_chain=common_pb.X509CertificateChain(
                certificates=chain
            )
        )


def _log_certificate_fingerprint(
    where: str, certificate: x509.Certificate, hash_algorithm: hashes.Hash
) -> None:
    """Log the fingerprint of a certificate, for debugging.

    Args:
        where: Location of where this gets called from, useful for debugging.
        certificate: Certificate to compute fingerprint of and log.
        hash_algorithm: The algorithm used to compute the fingerprint.
    """
    fp = certificate.fingerprint(hash_algorithm)
    logger.info(
        f"[{where:^8}] {hash_algorithm.name} "
        f"Fingerprint: {':'.join(f'{b:02X}' for b in fp)}"
    )


class Verifier(sigstore_pb.Verifier):
    """Verifier for signatures generated via signing with certificates."""

    def __init__(
        self,
        certificate_chain_paths: Iterable[pathlib.Path] = frozenset(),
        log_fingerprints: bool = False,
    ):
        """Initializes the verifier with the list of certificates to use.

        Args:
            certificate_chain_paths: Paths to certificates used to verify
              signature and establish chain of trust. By default this is empty,
              in which case we would use the root certificates from the
              operating system, as per `certifi.where()`.
            log_fingerprints: Log the fingerprints of certificates
        """
        self._log_fingerprints = log_fingerprints

        if not certificate_chain_paths:
            certificate_chain_paths = [pathlib.Path(p) for p in certifi.where()]

        certificates = x509.load_pem_x509_certificates(
            b"".join([path.read_bytes() for path in certificate_chain_paths])
        )

        self._store = crypto.X509Store()
        for certificate in certificates:
            if self._log_fingerprints:
                _log_certificate_fingerprint(
                    "init", certificate, hashes.SHA256()
                )
            self._store.add_cert(crypto.X509.from_cryptography(certificate))

    @override
    def _verify_bundle(self, bundle: bundle_pb.Bundle) -> tuple[str, bytes]:
        public_key = self._verify_certificates(bundle.verification_material)
        envelope = bundle.dsse_envelope
        try:
            public_key.verify(
                envelope.signatures[0].sig,
                sigstore_pb.pae(envelope.payload),
                ec.ECDSA(ec_key.get_ec_key_hash(public_key)),
            )
        except exceptions.InvalidSignature:
            # Compatibility layer with pre 1.0 release
            # Here, we patch over a bug in `pae` which mixed unicode `str` and
            # `bytes`. As a result, additional escape characters were added to
            # the material that got signed over.
            public_key.verify(
                envelope.signatures[0].sig,
                sigstore_pb.pae_compat(envelope.payload),
                # Note another bug here: the v0.2 signatures were generated with
                # hardcoded SHA256 hash, instead of the one that matches the
                # key type. To verify those signatures, we have to hardcode this
                # here too (instead of `ec_key.get_ec_key_hash(public_key)`).
                # For the hardcode path see:
                # https://github.com/sigstore/model-transparency/blob/9737f0e28349bf43897857ada7beaa22ec18e9a6/src/model_signing/signature/key.py#L103
                ec.ECDSA(hashes.SHA256()),
            )

        return envelope.payload_type, envelope.payload

    def _verify_certificates(
        self,
        verification_material: bundle_pb.VerificationMaterial,
        log_fingerprints: bool = False,
    ) -> ec.EllipticCurvePublicKey:
        """Verifies the certificate chain and returns the public key.

        The public key is extracted from the signing certificate from the chain
        of trust, after the chain is validated. It must match the public key
        from the key used during signing.
        """

        def _to_openssl_certificate(certificate_bytes, log_fingerprints):
            cert = x509.load_der_x509_certificate(certificate_bytes)
            if log_fingerprints:
                _log_certificate_fingerprint("verify", cert, hashes.SHA256())
            return crypto.X509.from_cryptography(cert)

        signing_chain = verification_material.x509_certificate_chain
        signing_certificate = x509.load_der_x509_certificate(
            signing_chain.certificates[0].raw_bytes
        )

        max_signing_time = signing_certificate.not_valid_before_utc
        self._store.set_time(max_signing_time)

        trust_chain_ssl = [
            _to_openssl_certificate(
                certificate.raw_bytes, self._log_fingerprints
            )
            for certificate in signing_chain.certificates[1:]
        ]
        signing_certificate_ssl = _to_openssl_certificate(
            signing_chain.certificates[0].raw_bytes, self._log_fingerprints
        )

        store_context = crypto.X509StoreContext(
            self._store, signing_certificate_ssl, trust_chain_ssl
        )
        store_context.verify_certificate()

        extensions = signing_certificate.extensions
        can_use_for_signing = False
        usage = extensions.get_extension_for_class(x509.KeyUsage)
        if usage.value.digital_signature:
            can_use_for_signing = True
        else:
            try:
                usage = extensions.get_extension_for_class(
                    x509.ExtendedKeyUsage
                )
                if oid.ExtendedKeyUsageOID.CODE_SIGNING in usage.value:
                    can_use_for_signing = True
            except x509.ExtensionNotFound:
                pass

        if not can_use_for_signing:
            raise ValueError("Signing certificate cannot be used for signing")

        return signing_certificate.public_key()



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Universal model signing library.

The API is split into 3 main components (and a glue `model_signing.manifest`
module for data types used in public interfaces):

- `model_signing.hashing`: responsible with generating a list of hashes for
  every component of the model. A component could be a file, a file shard, a
  tensor, etc., depending on the method used. We currently support only files
  and file shards. The result of hashing is a manifest, a listing of hashes for
  every object in the model.
- `model_signing.signing`: responsible with taking the manifest and generating a
  signature, based on a signing configuration. The signing configuration can
  select the method used to sign as well as the parameters.
- `model_signing.verifying`: responsible with taking a signature and verifying
  it. If the cryptographic parts of the signature can be validated, the
  verification layer would return an expanded manifest which can then be
  compared agains a manifest obtained from hashing the existing model. If the
  two manifest don't match then the model integrity was compromised and the
  `model_signing` package detected that.

The first two of these components allows configurability but can also be used
directly, with a default configuration. The only difference is for the
verification component where we need to configure the verification method since
there are no sensible defaults that can be used.

Signing can be done using the default configuration:

```python
model_signing.signing.sign("finbert", "finbert.sig")
```

This example generates the signature using Sigstore.

Alternatively, a custom configuration can be selected, for both signing and
hashing:

```python
model_signing.signing.Config().use_elliptic_key_signer(
    private_key="key"
).set_hashing_config(
    model_signing.hashing.Config().set_ignored_paths(
        paths=["README.md"], ignore_git_paths=True
    )
).sign("finbert", "finbert.sig")
```

This example generates a signature using a private key based on elliptic curve
cryptography. It also hashes the model by ignoring `README.md` and any git
related file present in the model directory.

We also support signing with signing certificates, using a similar API as above.

When verifying, we need to configure the cryptography configuration, so that the
code knows how to parse the signature.

For the Sigstore example, the simplest verification example would be:

```python
model_signing.verifying.Config().use_sigstore_verifier(
    identity=identity, oidc_issuer=oidc_provider
).verify("finbert", "finbert.sig")
```

Where `identity` and `oidc_provider` are the parameters obtained after the OIDC
flow during signing.

To verify the private key example, we could use the following:

```python
model_signing.verifying.Config().use_elliptic_key_verifier(
    public_key="key.pub"
).set_hashing_config(
    model_signing.hashing.Config().use_shard_serialization()
    )
).verify("finbert", "finbert.sig")
```

Alternatively, we also support automatic detection of the hashing configuration
during the verification process. So, the following should also work:

```python
model_signing.verifying.Config().use_elliptic_key_verifier(
    public_key="key.pub"
).verify("finbert", "finbert.sig")
```

A reminder that we still need to set the verification configuration. This sets
up the cryptographic primitives to verify the signature and is needed to know
how to parse the signature file.

For any signing method, the signature is a
[Sigstore bundle](https://docs.sigstore.dev/about/bundle/) which contains the
verification material (the information needed to verify the signature) and the
payload. The verification material depends on the method used for signing.

The payload in the signature is a
[DSSE envelope](https://github.com/secure-systems-lab/dsse) which contains an
[in-toto statement](https://github.com/in-toto/attestation/tree/main/spec/v1).
The in-toto statement contains the actual metadata that gets signed, and in our
case is a custom predicate that identifies all the components of the model.

Read more [on the repository's `README.md`][repo]. The CLI that maps over the
API is also documented there.

[repo]: https://github.com/sigstore/model-transparency/blob/main/README.md
"""

from model_signing import hashing
from model_signing import manifest
from model_signing import signing
from model_signing import verifying


__version__ = "1.0.1"


__all__ = ["hashing", "signing", "verifying", "manifest"]



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An in-memory serialized representation of an ML model.

A manifest pairs objects from the model (e.g., files, shards of files), with
their hashes. When signing the model we first generate a manifest from
serializing the model using a configured serialization method (see
`model_signing.signing`).

When verifying the integrity of the model, after checking the authenticity of
the signature, we extract a manifest from it. Then, we serialize the local model
(the model being tested) and compare the two manifests.

The serialization method used during signing must match the one used during
verification. We can auto detect the method to use during verification from the
signature, but it is recommended to be explicit when possible.

Comparing the manifests can be done by checking that every item matches, both in
name and in associated hash.  In the future we will support partial object
matching. This is useful, for example, for the cases where the original model
contained files for multiple ML frameworks, but the user only uses the model
with one framework. This way, the user can verify the integrity only for the
files that are actually used.

This API should not be used directly, we don't guarantee that it is fully stable
at the moment.
"""

import abc
from collections.abc import Iterable, Iterator
import dataclasses
import pathlib
import sys
from typing import Any, Final

from typing_extensions import override

from model_signing._hashing import hashing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclasses.dataclass(frozen=True, order=True)
class _ResourceDescriptor:
    """A description of any content from any `Manifest`.

    We aim this to be similar to in-toto's `ResourceDescriptor`. To support
    cases where in-toto cannot be directly used, we make this a dataclass that
    can be mapped to in-toto when needed, and used as its own otherwise.

    Not all fields from in-toto are specified at this moment. All fields here
    must be present, unlike in-toto, where all are optional.

    See github.com/in-toto/attestation/blob/main/spec/v1/resource_descriptor.md
    for the in-toto specification.

    Attributes:
        identifier: A string that uniquely identifies this object within the
        manifest. Depending on serialized format, users might require the
        identifier to be unique across all manifests stored in a system.
        Producers and consumers can agree on additional requirements (e.g.,
        several descriptors must have a common pattern for the identifier and
        the integrity of the model implies integrity of all these items,
        ignoring any other descriptor). Corresponds to `name`, `uri`, or
        `content` in in-toto specification.
        digest: One digest for the item. Note that unlike in-toto, we only have
          one digest for the item and it is always required.
    """

    identifier: str
    digest: hashing.Digest


class _ManifestKey(metaclass=abc.ABCMeta):
    """An object that can be a key for an item in a manifest.

    We need to be able to convert the key to string when we serialize the
    manifest and to rebuild the object from the serialized representation when
    we deserialize the manifest from the signature.
    """

    @classmethod
    @abc.abstractmethod
    def from_str(cls, s: str) -> Self:
        """Builds a manifest key from the string representation.

        It is guaranteed that for any `key` of a derived class `C` and a (valid)
        string `s`, the following two round-trip properties hold:

        ```
        str(C.from_str(s)) == s
        C.from_str(str(key)) == key
        ```

        Raises:
            ValueError: if the string argument cannot be decoded correctly.
        """


@dataclasses.dataclass(frozen=True, order=True)
class _File(_ManifestKey):
    """A dataclass to hold information about a file as a manifest key.

    Attributes:
        path: The path to the file, relative to the model root.
    """

    path: pathlib.PurePath

    def __str__(self) -> str:
        return str(self.path)

    @classmethod
    @override
    def from_str(cls, s: str) -> Self:
        # Note that we always decode the string to a pure POSIX path
        return cls(pathlib.PurePosixPath(s))


@dataclasses.dataclass(frozen=True, order=True)
class _Shard(_ManifestKey):
    """A dataclass to hold information about a file shard as a manifest key.

    Attributes:
        path: The path to the file, relative to the model root.
        start: The start offset of the shard (included).
        end: The end offset of the shard (not included).
    """

    path: pathlib.PurePath
    start: int
    end: int

    def __str__(self) -> str:
        return f"{str(self.path)}:{self.start}:{self.end}"

    @classmethod
    @override
    def from_str(cls, s: str) -> Self:
        parts = s.split(":")
        if len(parts) != 3:
            raise ValueError(f"Expected 3 components separated by `:`, got {s}")

        path = pathlib.PurePosixPath(parts[0])
        start = int(parts[1])
        end = int(parts[2])

        return cls(path, start, end)


class ManifestItem(metaclass=abc.ABCMeta):
    """An individual object of a model, stored as an item in a manifest.

    For example, this could be a file, or a file shard. All file paths are
    relative to the model root, to allow moving or renaming the model, without
    invalidating the signature.

    The integrity of each `ManifestItem` can be verified individually and in
    parallel. If the item is backed by a file, we recompute the hash of the
    portion of the file that represents this item.

    Attributes:
        digest: The digest of the item. Use the `key` property to obtain a
          canonical unique representation for the item.
    """

    digest: hashing.Digest

    @property
    @abc.abstractmethod
    def key(self) -> _ManifestKey:
        """A unique representation for the manifest item.

        Two items in the same manifest must not share the same `key`. The
        information contained in `key` should be sufficient to determine how to
        compute the item's digest.
        """


class FileManifestItem(ManifestItem):
    """A manifest item that records a filename path together with its digest.

    Note that the path component is a `pathlib.PurePath`, relative to the model.
    To ensure that the manifest is consistent across operating systems, we
    convert the path to a POSIX path.
    """

    def __init__(self, *, path: pathlib.PurePath, digest: hashing.Digest):
        """Builds a manifest item pairing a file with its digest.

        Args:
            path: The path to the file, relative to the model root.
            digest: The digest of the file.
        """
        # Note: we need to force a PurePosixPath to canonicalize the manifest.
        self._path = pathlib.PurePosixPath(path)
        self.digest = digest

    @property
    @override
    def key(self) -> _ManifestKey:
        return _File(self._path)


class ShardedFileManifestItem(ManifestItem):
    """A manifest item that records a file shard together with its digest.

    Note that the path component is a `pathlib.PurePath`, relative to the model.
    To ensure that the manifest is consistent across operating systems, we
    convert the path to a POSIX path.
    """

    def __init__(
        self,
        *,
        path: pathlib.PurePath,
        start: int,
        end: int,
        digest: hashing.Digest,
    ):
        """Builds a manifest item pairing a file shard with its digest.

        Args:
            path: The path to the file, relative to the model root.
            start: The start offset of the shard (included).
            end: The end offset of the shard (not included).
            digest: The digest of the file shard.
        """
        # Note: we need to force a PurePosixPath to canonicalize the manifest.
        self._path = pathlib.PurePosixPath(path)
        self._start = start
        self._end = end
        self.digest = digest

    @property
    @override
    def key(self) -> _ManifestKey:
        return _Shard(self._path, self._start, self._end)


class SerializationType(metaclass=abc.ABCMeta):
    """A description of the serialization process that generated the manifest.

    These should record all the parameters needed to ensure a reproducible
    serialization. These are used to build a manifest back from the signature in
    a backwards compatible way. We use these to determine what serialization to
    use when verifying a signature.
    """

    @property
    @abc.abstractmethod
    def serialization_parameters(self) -> dict[str, Any]:
        """The arguments of the serialization method."""

    @classmethod
    def from_args(cls, args: dict[str, Any]) -> Self:
        """Builds an instance of this class based on the dict representation.

        This is the reverse of `serialization_parameters`.

        Args:
            args: The arguments as a dictionary (equivalent to `**kwargs`).
        """
        serialization_type = args["method"]
        for subclass in [_FileSerialization, _ShardSerialization]:
            if serialization_type == subclass.method:
                return subclass._from_args(args)
        raise ValueError(f"Unknown serialization type {serialization_type}")

    @classmethod
    @abc.abstractmethod
    def _from_args(cls, args: dict[str, Any]) -> Self:
        """Performs the actual build from `from_dict`."""

    @abc.abstractmethod
    def new_item(self, name: str, digest: hashing.Digest) -> ManifestItem:
        """Builds a `ManifestItem` of the correct type.

        Each serialization type results in different types for the items in the
        manifest. This method parses the `name` of the item according to the
        serialization type to create the proper manifest item.

        Args:
            name: The name of the item, as shown in the manifest.
            digest: The digest of the item.
        """


class _FileSerialization(SerializationType):
    method: Final[str] = "files"

    def __init__(self, hash_type: str, allow_symlinks: bool = False):
        """Records the manifest serialization type for serialization by files.

        We only need to record the hashing engine used and whether symlinks are
        hashed or ignored.

        Args:
            hash_type: A string representation of the hash algorithm.
            allow_symlinks: Controls whether symbolic links are included.
        """
        self._hash_type = hash_type
        self._allow_symlinks = allow_symlinks

    @property
    @override
    def serialization_parameters(self) -> dict[str, Any]:
        return {
            "method": self.method,
            "hash_type": self._hash_type,
            "allow_symlinks": self._allow_symlinks,
        }

    @classmethod
    @override
    def _from_args(cls, args: dict[str, Any]) -> Self:
        return cls(args["hash_type"], args["allow_symlinks"])

    @override
    def new_item(self, name: str, digest: hashing.Digest) -> ManifestItem:
        path = pathlib.PurePosixPath(name)
        return FileManifestItem(path=path, digest=digest)


class _ShardSerialization(SerializationType):
    method: Final[str] = "shards"

    def __init__(
        self, hash_type: str, shard_size: int, allow_symlinks: bool = False
    ):
        """Records the manifest serialization type for serialization by files.

        We need to record the hashing engine used and whether symlinks are
        hashed or ignored, just like for file serialization. We also need to
        record the shard size used to split the files, since different shard
        sizes results in different resources.

        Args:
            hash_type: A string representation of the hash algorithm.
            allow_symlinks: Controls whether symbolic links are included.
        """
        self._hash_type = hash_type
        self._allow_symlinks = allow_symlinks
        self._shard_size = shard_size

    @property
    @override
    def serialization_parameters(self) -> dict[str, Any]:
        return {
            "method": self.method,
            "hash_type": self._hash_type,
            "shard_size": self._shard_size,
            "allow_symlinks": self._allow_symlinks,
        }

    @classmethod
    @override
    def _from_args(cls, args: dict[str, Any]) -> Self:
        return cls(
            args["hash_type"], args["shard_size"], args["allow_symlinks"]
        )

    @override
    def new_item(self, name: str, digest: hashing.Digest) -> ManifestItem:
        parts = name.split(":")
        if len(parts) != 3:
            raise ValueError(
                "Invalid resource name: expected 3 components separated by "
                f"`:`, got {name}"
            )

        path = pathlib.PurePosixPath(parts[0])
        start = int(parts[1])
        end = int(parts[2])
        return ShardedFileManifestItem(
            path=path, start=start, end=end, digest=digest
        )


class Manifest:
    """Generic manifest file to represent a model."""

    def __init__(
        self,
        model_name: str,
        items: Iterable[ManifestItem],
        serialization_type: SerializationType,
    ):
        """Builds a manifest from a collection of already hashed objects.

        Args:
            model_name: A name for the model that generated the manifest. This
              is the final component of the model path, and is only informative.
              See `model_name` property.
            items: An iterable sequence of objects and their hashes.
        """
        self._name = model_name
        self._item_to_digest = {item.key: item.digest for item in items}
        self._serialization_type = serialization_type

    def __eq__(self, other: Self):
        return self._item_to_digest == other._item_to_digest

    def resource_descriptors(self) -> Iterator[_ResourceDescriptor]:
        """Yields each resource from the manifest, one by one."""
        for item, digest in sorted(self._item_to_digest.items()):
            yield _ResourceDescriptor(identifier=str(item), digest=digest)

    @property
    def model_name(self) -> str:
        """The name of the model when serialized (final component of the path).

        This is only informative. Changing the name of the model should still
        result in the same digests after serialization, it must not invalidate
        signatures. As a result, two manifests with different model names but
        with the same resource descriptors will compare equal.
        """
        return self._name

    @property
    def serialization_type(self) -> dict[str, Any]:
        """The serialization (and arguments) used to build the manifest.

        This is needed to record the serialization method used to generate the
        manifest so that signature verification can use the same method.
        """
        return self._serialization_type.serialization_parameters



# Copyright 2025 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The main entry-point for the model_signing package."""

from collections.abc import Iterable, Sequence
import logging
import pathlib
import sys
from typing import Optional

import click

import model_signing


logging.basicConfig(format="%(message)s", level=logging.INFO)


# Decorator for the commonly used argument for the model path.
_model_path_argument = click.argument(
    "model_path", type=pathlib.Path, metavar="MODEL_PATH"
)


# Decorator for the commonly used option to set the signature path when signing.
_write_signature_option = click.option(
    "--signature",
    type=pathlib.Path,
    metavar="SIGNATURE_PATH",
    default=pathlib.Path("model.sig"),
    help="Location of the signature file to generate. Defaults to `model.sig`.",
)


# Decorator for the commonly used option for the signature to verify.
_read_signature_option = click.option(
    "--signature",
    type=pathlib.Path,
    metavar="SIGNATURE_PATH",
    required=True,
    help="Location of the signature file to verify.",
)

# Decorator for the commonly used option to ignore certain paths
_ignore_paths_option = click.option(
    "--ignore-paths",
    type=pathlib.Path,
    metavar="IGNORE_PATHS",
    multiple=True,
    help="File paths to ignore when signing or verifying.",
)

# Decorator for the commonly used option to ignore git-related paths
_ignore_git_paths_option = click.option(
    "--ignore-git-paths/--no-ignore-git-paths",
    type=bool,
    default=True,
    show_default=True,
    help="Ignore git-related files when signing or verifying.",
)

# Decorator for the commonly used option to set the path to the private key
# (when using non-Sigstore PKI).
_private_key_option = click.option(
    "--private_key",
    type=pathlib.Path,
    metavar="PRIVATE_KEY",
    required=True,
    help="Path to the private key, as a PEM-encoded file.",
)


# Decorator for the commonly used option to pass a certificate chain to
# establish root of trust (when signing or verifying using certificates).
_certificate_root_of_trust_option = click.option(
    "--certificate_chain",
    type=pathlib.Path,
    metavar="CERTIFICATE_PATH",
    multiple=True,
    help="Path to certificate chain of trust.",
)


# Decorator for the commonly used option to use Sigstore's staging instance.
_sigstore_staging_option = click.option(
    "--use_staging",
    type=bool,
    is_flag=True,
    help="Use Sigstore's staging instance.",
)


class _PKICmdGroup(click.Group):
    """A custom group to configure the supported PKI methods."""

    _supported_modes = ["sigstore", "key", "certificate"]

    def get_command(
        self, ctx: click.Context, cmd_name: str
    ) -> Optional[click.Command]:
        """Retrieves a command with a given name.

        We use this to make Sigstore signing be the default, if it is missing.
        """
        if cmd_name in self._supported_modes:
            return super().get_command(ctx, cmd_name)
        return super().get_command(ctx, "sigstore")

    def resolve_command(
        self, ctx: click.Context, args: Sequence[str]
    ) -> tuple[Optional[str], Optional[click.Command], Iterable[str]]:
        """Resolves a command and its arguments.

        We use this to make Sigstore signing be the default and correctly alter
        the arguments. We are guaranteed that `args` has at least one element
        (otherwise the help menu would be printed). This argument should be the
        subcommand and would be removed as a result of this function, in
        general.

        However, if the first argument does not resolve to a supported PKI
        method, then we inject "sigstore" as the subcommand (in `get_command`).
        All that is left to do is to pass all `args` to the subcommand, without
        removing anything.
        """
        if args[0] in self._supported_modes:
            return super().resolve_command(ctx, args)
        _, cmd, _ = super().resolve_command(ctx, args)
        return cmd.name, cmd, args


@click.group(
    context_settings=dict(help_option_names=["-h", "--help"]),
    epilog=(
        "Check https://sigstore.github.io/model-transparency for "
        "documentation and more details."
    ),
)
@click.version_option(model_signing.__version__, "--version")
def main() -> None:
    """ML model signing and verification.

    Use each subcommand's `--help` option for details on each mode.
    """


@main.group(name="sign", subcommand_metavar="PKI_METHOD", cls=_PKICmdGroup)
def _sign() -> None:
    """Sign models.

    Produces a cryptographic signature (in the form of a Sigstore bundle) for a
    model. We support any model format, either as a single file or as a
    directory.

    We support multiple PKI methods, specified as subcommands. By default, the
    signature is generated via Sigstore (as if invoking `sigstore` subcommand).

    Use each subcommand's `--help` option for details on each mode.
    """


@_sign.command(name="sigstore")
@_model_path_argument
@_ignore_paths_option
@_ignore_git_paths_option
@_write_signature_option
@_sigstore_staging_option
@click.option(
    "--use_ambient_credentials",
    type=bool,
    is_flag=True,
    help="Use credentials from ambient environment.",
)
@click.option(
    "--identity_token",
    type=str,
    metavar="TOKEN",
    help=(
        "Fixed OIDC identity token to use instead of obtaining credentials "
        "from OIDC flow or from the environment."
    ),
)
def _sign_sigstore(
    model_path: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    signature: pathlib.Path,
    use_ambient_credentials: bool,
    use_staging: bool,
    identity_token: Optional[str] = None,
) -> None:
    """Sign using Sigstore (DEFAULT signing method).

    Signing the model at MODEL_PATH, produces the signature at SIGNATURE_PATH
    (as per `--signature` option). Files in IGNORE_PATHS are not part of the
    signature.

    If using Sigstore, we need to provision an OIDC token. In general, this is
    taken from an interactive OIDC flow, but ambient credentials could be used
    to use workload identity tokens (e.g., when running in GitHub actions).
    Alternatively, a constant identity token can be provided via
    `--identity_token`.

    Sigstore allows users to use a staging instance for test-only signatures.
    Passing the `--use_staging` flag would use that instance instead of the
    production one.
    """
    try:
        model_signing.signing.Config().use_sigstore_signer(
            use_ambient_credentials=use_ambient_credentials,
            use_staging=use_staging,
            identity_token=identity_token,
        ).set_hashing_config(
            model_signing.hashing.Config().set_ignored_paths(
                paths=list(ignore_paths) + [signature],
                ignore_git_paths=ignore_git_paths,
            )
        ).sign(model_path, signature)
    except Exception as err:
        click.echo(f"Signing failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Signing succeeded")


@_sign.command(name="key")
@_model_path_argument
@_ignore_paths_option
@_ignore_git_paths_option
@_write_signature_option
@_private_key_option
@click.option(
    "--password",
    type=str,
    metavar="PASSWORD",
    help="Password for the key encryption, if any",
)
def _sign_private_key(
    model_path: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    signature: pathlib.Path,
    private_key: pathlib.Path,
    password: Optional[str] = None,
) -> None:
    """Sign using a private key (paired with a public one).

    Signing the model at MODEL_PATH, produces the signature at SIGNATURE_PATH
    (as per `--signature` option). Files in IGNORE_PATHS are not part of the
    signature.

    Traditionally, signing could be achieved by using a public/private key pair.
    Pass the signing key using `--private_key`.

    Note that this method does not provide a way to tie to the identity of the
    signer, outside of pairing the keys. Also note that we don't offer key
    management protocols.
    """
    try:
        model_signing.signing.Config().use_elliptic_key_signer(
            private_key=private_key, password=password
        ).set_hashing_config(
            model_signing.hashing.Config().set_ignored_paths(
                paths=list(ignore_paths) + [signature],
                ignore_git_paths=ignore_git_paths,
            )
        ).sign(model_path, signature)
    except Exception as err:
        click.echo(f"Signing failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Signing succeeded")


@_sign.command(name="certificate")
@_model_path_argument
@_ignore_paths_option
@_ignore_git_paths_option
@_write_signature_option
@_private_key_option
@click.option(
    "--signing_certificate",
    type=pathlib.Path,
    metavar="CERTIFICATE_PATH",
    required=True,
    help="Path to the signing certificate, as a PEM-encoded file.",
)
@_certificate_root_of_trust_option
def _sign_certificate(
    model_path: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    signature: pathlib.Path,
    private_key: pathlib.Path,
    signing_certificate: pathlib.Path,
    certificate_chain: Iterable[pathlib.Path],
) -> None:
    """Sign using a certificate.

    Signing the model at MODEL_PATH, produces the signature at SIGNATURE_PATH
    (as per `--signature` option). Files in IGNORE_PATHS are not part of the
    signature.

    Traditionally, signing can be achieved by using keys from a certificate.
    The certificate can also provide the identity of the signer, making this
    method more informative than just using a public/private key pair for
    signing.  Pass the private signing key using `--private_key` and signing
    certificate via `--signing_certificate`. Optionally, pass a certificate
    chain via `--certificate_chain` to establish root of trust (this option can
    be repeated as needed, or all cerificates could be placed in a single file).

    Note that we don't offer certificate and key management protocols.
    """
    try:
        model_signing.signing.Config().use_certificate_signer(
            private_key=private_key,
            signing_certificate=signing_certificate,
            certificate_chain=certificate_chain,
        ).set_hashing_config(
            model_signing.hashing.Config().set_ignored_paths(
                paths=list(ignore_paths) + [signature],
                ignore_git_paths=ignore_git_paths,
            )
        ).sign(model_path, signature)
    except Exception as err:
        click.echo(f"Signing failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Signing succeeded")


@main.group(name="verify", subcommand_metavar="PKI_METHOD", cls=_PKICmdGroup)
def _verify() -> None:
    """Verify models.

    Given a model and a cryptographic signature (in the form of a Sigstore
    bundle) for the model, this call checks that the model matches the
    signature, that the model has not been tampered with. We support any model
    format, either as a signle file or as a directory.

    We support multiple PKI methods, specified as subcommands. By default, the
    signature is assumed to be generated via Sigstore (as if invoking `sigstore`
    subcommand).

    Use each subcommand's `--help` option for details on each mode.
    """


@_verify.command(name="sigstore")
@_model_path_argument
@_read_signature_option
@_ignore_paths_option
@_ignore_git_paths_option
@_sigstore_staging_option
@click.option(
    "--identity",
    type=str,
    metavar="IDENTITY",
    required=True,
    help="The expected identity of the signer (e.g., name@example.com).",
)
@click.option(
    "--identity_provider",
    type=str,
    metavar="IDENTITY_PROVIDER",
    required=True,
    help="The expected identity provider (e.g., https://accounts.example.com).",
)
def _verify_sigstore(
    model_path: pathlib.Path,
    signature: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    identity: str,
    identity_provider: str,
    use_staging: bool,
) -> None:
    """Verify using Sigstore (DEFAULT verification method).

    Verifies the integrity of model at MODEL_PATH, according to signature from
    SIGNATURE_PATH (given via `--signature` option). Files in IGNORE_PATHS are
    ignored.

    For Sigstore, we also need to provide an expected identity and identity
    provider for the signature. If these don't match what is provided in the
    signature, verification would fail.
    """
    try:
        model_signing.verifying.Config().use_sigstore_verifier(
            identity=identity,
            oidc_issuer=identity_provider,
            use_staging=use_staging,
        ).set_hashing_config(
            model_signing.hashing.Config().set_ignored_paths(
                paths=list(ignore_paths) + [signature],
                ignore_git_paths=ignore_git_paths,
            )
        ).verify(model_path, signature)
    except Exception as err:
        click.echo(f"Verification failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Verification succeeded")


@_verify.command(name="key")
@_model_path_argument
@_read_signature_option
@_ignore_paths_option
@_ignore_git_paths_option
@click.option(
    "--public_key",
    type=pathlib.Path,
    metavar="PUBLIC_KEY",
    required=True,
    help="Path to the public key used for verification.",
)
def _verify_private_key(
    model_path: pathlib.Path,
    signature: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    public_key: pathlib.Path,
) -> None:
    """Verity using a public key (paired with a private one).

    Verifies the integrity of model at MODEL_PATH, according to signature from
    SIGNATURE_PATH (given via `--signature` option). Files in IGNORE_PATHS are
    ignored.

    The public key provided via `--public_key` must have been paired with the
    private key used when generating the signature.

    Note that this method does not provide a way to tie to the identity of the
    signer, outside of pairing the keys. Also note that we don't offer key
    management protocols.
    """
    try:
        model_signing.verifying.Config().use_elliptic_key_verifier(
            public_key=public_key
        ).set_hashing_config(
            model_signing.hashing.Config().set_ignored_paths(
                paths=list(ignore_paths) + [signature],
                ignore_git_paths=ignore_git_paths,
            )
        ).verify(model_path, signature)
    except Exception as err:
        click.echo(f"Verification failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Verification succeeded")


@_verify.command(name="certificate")
@_model_path_argument
@_read_signature_option
@_ignore_paths_option
@_ignore_git_paths_option
@_certificate_root_of_trust_option
@click.option(
    "--log_fingerprints",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help="Log SHA256 fingerprints of all certificates.",
)
def _verify_certificate(
    model_path: pathlib.Path,
    signature: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    certificate_chain: Iterable[pathlib.Path],
    log_fingerprints: bool,
) -> None:
    """Verify using a certificate.

    Verifies the integrity of model at MODEL_PATH, according to signature from
    SIGNATURE_PATH (given via `--signature` option). Files in IGNORE_PATHS are
    ignored.

    The signing certificate is encoded in the signature, as part of the Sigstore
    bundle. To verify the root of trust, pass additional certificates in the
    certificate chain, using `--certificate_chain` (this option can be repeated
    as needed, or all certificates could be placed in a single file).

    Note that we don't offer certificate and key management protocols.
    """
    try:
        model_signing.verifying.Config().use_certificate_verifier(
            certificate_chain=certificate_chain,
            log_fingerprints=log_fingerprints,
        ).set_hashing_config(
            model_signing.hashing.Config().set_ignored_paths(
                paths=list(ignore_paths) + [signature],
                ignore_git_paths=ignore_git_paths,
            )
        ).verify(model_path, signature)
    except Exception as err:
        click.echo(f"Verification failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Verification succeeded")



# Copyright 2025 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Signers and verifiers using elliptic curve keys."""

import pathlib
from typing import Optional

from cryptography import exceptions
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from google.protobuf import json_format
from sigstore_protobuf_specs.dev.sigstore.bundle import v1 as bundle_pb
from sigstore_protobuf_specs.dev.sigstore.common import v1 as common_pb
from sigstore_protobuf_specs.io import intoto as intoto_pb
from typing_extensions import override

from model_signing._signing import sign_sigstore_pb as sigstore_pb
from model_signing._signing import signing


def _check_supported_ec_key(public_key: ec.EllipticCurvePublicKey):
    """Checks if the elliptic curve key is supported by our package.

    We only support a family of curves, trying to match those specified by
    Sigstore's protobuf specs.
    See https://github.com/sigstore/model-transparency/issues/385.

    Args:
        public_key: The public key to check. Can be obtained from a private key.

    Raises:
        ValueError: The key is not supported.
    """
    curve = public_key.curve.name
    if curve not in ["secp256r1", "secp384r1", "secp521r1"]:
        raise ValueError(f"Unsupported key for curve '{curve}'")


def get_ec_key_hash(
    public_key: ec.EllipticCurvePublicKey,
) -> hashes.HashAlgorithm:
    """Returns the public key hashing algorithm.

    We need to record this in the sigstore bundle when signing and retrieve when
    performing verification. This is according to sigstore protobuf specs and is
    used both when signing with keys and when signing with certificates.

    Args:
        public_key: The public key to get the hash algorithm from.

    Raises:
        ValueError: The key is not supported.
    """
    key_size = public_key.curve.key_size

    # TODO: Once Python 3.9 support is deprecated revert to using `match`
    if key_size == 256:
        return hashes.SHA256()
    if key_size == 384:
        return hashes.SHA384()
    if key_size == 521:
        return hashes.SHA512()

    raise ValueError(f"Unexpected key size {key_size}")


class Signer(sigstore_pb.Signer):
    """Signer using an elliptic curve private key."""

    def __init__(
        self, private_key_path: pathlib.Path, password: Optional[str] = None
    ):
        """Initializes the signer with the private key and optional password.

        Args:
            private_key_path: The path to the PEM encoded private key.
            password: Optional password for the private key.
        """
        self._private_key = serialization.load_pem_private_key(
            private_key_path.read_bytes(), password
        )
        _check_supported_ec_key(self._private_key.public_key())

    @override
    def sign(self, payload: signing.Payload) -> signing.Signature:
        raw_payload = json_format.MessageToJson(payload.statement.pb).encode(
            "utf-8"
        )

        raw_signature = intoto_pb.Signature(
            sig=self._private_key.sign(
                sigstore_pb.pae(raw_payload),
                ec.ECDSA(get_ec_key_hash(self._private_key.public_key())),
            ),
            keyid=None,
        )

        envelope = intoto_pb.Envelope(
            payload=raw_payload,
            payload_type=signing._IN_TOTO_JSON_PAYLOAD_TYPE,
            signatures=[raw_signature],
        )

        return sigstore_pb.Signature(
            bundle_pb.Bundle(
                media_type=sigstore_pb._BUNDLE_MEDIA_TYPE,
                verification_material=self._get_verification_material(),
                dsse_envelope=envelope,
            )
        )

    def _get_verification_material(self) -> bundle_pb.VerificationMaterial:
        """Returns the verification material to include in the bundle."""
        public_key = self._private_key.public_key()
        key_size = public_key.curve.key_size

        # TODO: Once Python 3.9 support is deprecated revert to using `match`
        if key_size == 256:
            key_details = common_pb.PublicKeyDetails.PKIX_ECDSA_P256_SHA_256
        elif key_size == 384:
            key_details = common_pb.PublicKeyDetails.PKIX_ECDSA_P384_SHA_384
        elif key_size == 521:
            key_details = common_pb.PublicKeyDetails.PKIX_ECDSA_P521_SHA_512
        else:
            raise ValueError(f"Unexpected key size {key_size}")

        return bundle_pb.VerificationMaterial(
            public_key=common_pb.PublicKey(
                raw_bytes=public_key.public_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PublicFormat.SubjectPublicKeyInfo,
                ),
                key_details=key_details,
            )
        )


class Verifier(sigstore_pb.Verifier):
    """Verifier for signatures generated with an elliptic curve private key."""

    def __init__(self, public_key_path: pathlib.Path):
        """Initializes the verifier with the public key to use.

        Args:
            public_key_path: The path to the public key to use. This must be
              paired with the private key used to generate the signature.
        """
        self._public_key = serialization.load_pem_public_key(
            public_key_path.read_bytes()
        )
        _check_supported_ec_key(self._public_key)

    @override
    def _verify_bundle(self, bundle: bundle_pb.Bundle) -> tuple[str, bytes]:
        # TODO: This method currently does not verify that the public key
        # matches what is included in the signature's verification material.
        # Instead, the verification material is completely ignored right now.

        envelope = bundle.dsse_envelope
        try:
            self._public_key.verify(
                envelope.signatures[0].sig,
                sigstore_pb.pae(envelope.payload),
                ec.ECDSA(get_ec_key_hash(self._public_key)),
            )
        except exceptions.InvalidSignature:
            # Compatibility layer with pre 1.0 release
            # Here, we patch over a bug in `pae` which mixed unicode `str` and
            # `bytes`. As a result, additional escape characters were added to
            # the material that got signed over.
            self._public_key.verify(
                envelope.signatures[0].sig,
                sigstore_pb.pae_compat(envelope.payload),
                ec.ECDSA(get_ec_key_hash(self._public_key)),
            )

        return envelope.payload_type, envelope.payload



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machinery for signing and verification of ML models.

The serialization API produces a manifest representation of the models, and we
use that to implement integrity checking of models in different computational
patterns. This means that all manifests need to be kept only in memory.

Since we need to support multiple signing methods (e.g., Sigstore, key,
certificate, etc.) , we provide a `Signer` abstract class with a single `sign`
method that takes a signing payload and converts it to a signature in the
supported format.

Finally, every signature needs to be verified. We pair every `Signer` subclass
with a `Verifier` which takes a signature, verify the authenticity of the
payload and then expand that to a manifest.

Regarding the data formats, for signing, we need to convert the manifest to the
signing payload. We only support manifests serialized to in-toto formats
described by https://github.com/in-toto/attestation/tree/main/spec/v1. The
envelope format is DSSE, as described in
https://github.com/secure-systems-lab/dsse. The signature is in a Sigstore
bundle format over the DSSE payload. The format is described at
https://docs.sigstore.dev/about/bundle/.
"""

import abc
import json
import pathlib
import sys
from typing import Any

from in_toto_attestation.v1 import statement

from model_signing import manifest
from model_signing._hashing import hashing
from model_signing._hashing import memory


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


# The expected in-toto payload type for the signature.
_IN_TOTO_JSON_PAYLOAD_TYPE: str = "application/vnd.in-toto+json"


# The expected in-toto statement type for the signature.
_IN_TOTO_STATEMENT_TYPE: str = "https://in-toto.io/Statement/v1"


# The expected model signature predicate type.
_PREDICATE_TYPE: str = "https://model_signing/signature/v1.0"


# The expected model signature predicate type for v0.2 (compat)
_PREDICATE_TYPE_COMPAT: str = "https://model_signing/Digests/v0.1"


def dsse_payload_to_manifest(dsse_payload: dict[str, Any]) -> manifest.Manifest:
    """Builds a manifest from the DSSE payload read from a signature.

    The payload here is a dictionary that represents the payload part of the
    DSSE envelope contained in the Sigstore bundle.

    Args:
        payload: The in-toto DSSE envelope to convert to manifest.

    Returns:
        A manifest representing the signed model.

    Raises:
        ValueError: The payload cannot be deserialized to a manifest.
    """
    obtained_predicate_type = dsse_payload["predicateType"]
    if obtained_predicate_type != _PREDICATE_TYPE:
        if obtained_predicate_type == _PREDICATE_TYPE_COMPAT:
            return dsse_payload_to_manifest_compat(dsse_payload)
        raise ValueError(
            f"Predicate type mismatch, expected {_PREDICATE_TYPE}, "
            f"got {obtained_predicate_type}"
        )

    subjects = dsse_payload["subject"]
    if len(subjects) != 1:
        raise ValueError(f"Expected only one subject, got {subjects}")

    model_name = subjects[0]["name"]
    expected_digest = subjects[0]["digest"]["sha256"]

    predicate = dsse_payload["predicate"]
    serialization_args = predicate["serialization"]
    serialization = manifest.SerializationType.from_args(serialization_args)

    hasher = memory.SHA256()
    items = []
    for resource in predicate["resources"]:
        name = resource["name"]
        algorithm = resource["algorithm"]
        digest_value = resource["digest"]
        digest = hashing.Digest(algorithm, bytes.fromhex(digest_value))
        hasher.update(digest.digest_value)
        items.append(serialization.new_item(name, digest))

    obtained_digest = hasher.compute().digest_hex
    if obtained_digest != expected_digest:
        raise ValueError(
            f"Manifest is inconsistent. Root digest is {expected_digest}, "
            f"but the included resources hash to {obtained_digest}"
        )

    return manifest.Manifest(model_name, items, serialization)


def dsse_payload_to_manifest_compat(
    dsse_payload: dict[str, Any],
) -> manifest.Manifest:
    """Builds a manifest from the DSSE payload read from a signature.

    This is the same as `dsse_payload_to_manifest` but using a DSSE payload as
    defined at v0.2 release. This was experimental but got used in production
    before v1.0 so we need to patch support to it while verifiers migrate to the
    forward compatible format defined by v1.0.

    Args:
        payload: The in-toto DSSE envelope to convert to manifest.

    Returns:
        A manifest representing the signed model.

    Raises:
        ValueError: The payload cannot be deserialized to a manifest.
    """
    # Model name is not defined, use a constant.
    model_name = "compat-undefined-not-present"

    # Serialization format is not present, build a fake one.
    serialization = manifest.SerializationType.from_args(
        {"method": "files", "hash_type": "sha256", "allow_symlinks": "false"}
    )

    # The only field with actual content is the subject.
    items = []
    for subject in dsse_payload["subject"]:
        name = subject["name"]
        algorithm = "sha256"  # hardcoded, the only supported one
        digest_value = subject["digest"][algorithm]
        digest = hashing.Digest(algorithm, bytes.fromhex(digest_value))
        items.append(serialization.new_item(name, digest))

    # There is no verification that the manifest is missing items at this point.
    return manifest.Manifest(model_name, items, serialization)


class Payload:
    """In-toto payload used to represent a model for signing.

    This payload represents all the object (files, shards, etc.) of the model
    paired with their hashes. It can be seen as a serialization of a manifest.
    The hashes are all recorded under the predicate, given that for the subject
    we are limited on what hashes we can use
    (https://github.com/sigstore/sigstore-python/issues/1018). Each hash follows
    the format of a ResourceDescriptor: is an object containing a name for the
    object, the hashing algorithm, and the digest value. These are recorded in
    the predicate, as part of the `"resources"` list.

    The subject is a name for the model (taken from the model's directory) and a
    global digest over all the computed digests. This is SHA256 computed over
    all the digests, in the order they show up in the predicate (we canonicalize
    this to be in alphabetical order). This digest can be used to refer to the
    model from other metadata documents without having to carry the entire set
    of resource descriptors around.

    To ensure backwards compatibility, the predicate contains a
    `"serialization"` section which describes the method used to serialize a
    model to the manifest used to generate this payload. The section includes a
    method name and a list of all relevant values needed to recompute the
    serialization.

    Future extensions to the model signature (e.g., incorporating model cards,
    etc.) can be added as part of the predicate. For v1.0 of the predicate the
    only supported fields in the predicate are `"serialization"` and
    `"resources"`. Any other field should be ignored by verifiers adhering to
    v1.0 version.

    Example:
    ```json
    {
      "_type": "https://in-toto.io/Statement/v1",
      "subject": [
        {
          "name": "sample_model",
          "digest": {
            "sha256": "143cc6..."
          }
        }
      ],
      "predicateType": "https://model_signing/signature/v1.0",
      "predicate": {
        "serialization": {
          "method": "files",
          "hash_type": "sha256",
          "allow_symlinks": true
        },
        "resources": [
          {
            "algorithm": "sha256",
            "digest": "fdd892...",
            "name": "d0/f00"
          },
          {
            "algorithm": "sha256",
            "digest": "e16940...",
            "name": "d0/f01"
          },
          {
            "algorithm": "sha256",
            "digest": "407822...",
            "name": "d0/f02"
          },
          ...
          {
            "algorithm": "sha256",
            "digest": "912bcf...",
            "name": "f3"
          }
        ]
      }
    }
    ```
    """

    def __init__(self, manifest: manifest.Manifest):
        """Builds an instance of this in-toto payload.

        Args:
            manifest: the manifest to convert to signing payload.
        """
        hasher = memory.SHA256()
        resources = []
        for descriptor in manifest.resource_descriptors():
            hasher.update(descriptor.digest.digest_value)
            resources.append(
                {
                    "name": descriptor.identifier,
                    "algorithm": descriptor.digest.algorithm,
                    "digest": descriptor.digest.digest_hex,
                }
            )

        root_digest = {"sha256": hasher.compute().digest_hex}
        subject = statement.ResourceDescriptor(
            name=manifest.model_name, digest=root_digest
        ).pb

        predicate = {
            "serialization": manifest.serialization_type,
            "resources": resources,
            # other properties can go here
        }

        self.statement = statement.Statement(
            subjects=[subject],
            predicate_type=_PREDICATE_TYPE,
            predicate=predicate,
        )


class Signature(metaclass=abc.ABCMeta):
    """Signature class, wrapping a sigstore bundle.

    We only support sigstore bundle signature formats, but we need to have two
    separate classes for this given the need to support traditional signing as
    well as Sigstore one. One class wraps around `sigstore_models.Bundle` and
    the other around the bundle as defined by `sigstore_protobuf_specs`.
    """

    @abc.abstractmethod
    def write(self, path: pathlib.Path) -> None:
        """Writes the signature to disk, to the given path.

        Args:
            path: The path to write the signature to.
        """

    @classmethod
    @abc.abstractmethod
    def read(cls, path: pathlib.Path) -> Self:
        """Reads the signature from disk.

        Does not perform any signature verification, except what is needed to
        parse the signature file.

        Args:
            path: The path to read the signature from.

        Returns:
            An instance of the class which can be passed to a `Verifier` for
            signature and integrity verification.
        """


class Signer(metaclass=abc.ABCMeta):
    """Generic signer.

    Each signer may implement its own mechanism for managing the key material.
    """

    @abc.abstractmethod
    def sign(self, payload: Payload) -> Signature:
        """Signs the provided signing payload.

        Args:
            payload: The `Payload` instance that should be signed.

        Returns:
            A valid signature.
        """


class Verifier(metaclass=abc.ABCMeta):
    """Generic signature verifier.

    Every subclass of `Verifier` is paired with a subclass of `Signer`. This is
    to ensure that they support the same signature formats as well as have
    similar key materials.

    If the signature is valid, the payload is expanded to a `Manifest` instance
    which can then be used to check the model integrity.
    """

    def verify(self, signature: Signature) -> manifest.Manifest:
        """Verifies the signature.

        Args:
            signature: The signature to verify.

        Returns:
            A `manifest.Manifest` instance that represents the model.

        Raises:
            ValueError: Signature verification fails.
        """
        payload_type, payload = self._verify_signed_content(signature)

        if payload_type != _IN_TOTO_JSON_PAYLOAD_TYPE:
            raise ValueError(
                f"Expected DSSE payload {_IN_TOTO_JSON_PAYLOAD_TYPE}, "
                f"but got {payload_type}"
            )

        payload = json.loads(payload)

        if payload["_type"] != _IN_TOTO_STATEMENT_TYPE:
            raise ValueError(
                f"Expected in-toto {_IN_TOTO_STATEMENT_TYPE} payload, "
                f"but got {payload['_type']}"
            )

        return dsse_payload_to_manifest(payload)

    @abc.abstractmethod
    def _verify_signed_content(self, signature: Signature) -> tuple[str, bytes]:
        """Verifies the signed content and extract payload type and payload.

        Subclasses only need to implement this method.

        Args:
            signature: The signature to verify.

        Returns:
            A tuple containing the payload type and the payload (as a JSON
            object loaded to a dictionary).
        """
